USE
	TempTablesDemoDB;
GO


SET STATISTICS IO ON;
GO


-- Create the "Marketing.GetLastThreeOrdersForSelectedCustomers" stored procedure with a table variable

CREATE OR ALTER PROCEDURE
	Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
(
	@NumberOfCustomers AS INT
)
AS

DECLARE
	@SelectedCustomers
TABLE
(
	CustomerId INT NOT NULL
);

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

SELECT
	CustomerId		= SelectedCustomers.CustomerId ,
	OrderId			= LastThreeOrders.OrderId ,
	OrderDateTime	= LastThreeOrders.OrderDateTime ,
	OrderStatusId	= LastThreeOrders.OrderStatusId
FROM
	@SelectedCustomers AS SelectedCustomers
OUTER APPLY
	(
		SELECT TOP (3)
			OrderId			= Orders.Id ,
			OrderDateTime	= Orders.DateAndTime ,
			OrderStatusId	= Orders.OrderStatusId
		FROM
			Sales.Orders AS Orders
		WHERE
			Orders.CustomerId = SelectedCustomers.CustomerId
		ORDER BY
			OrderDateTime DESC
	)
	AS
		LastThreeOrders;
GO


-- Test the stored procedure
-- Notice that the estimated number of rows in the table variable is 1

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
	@NumberOfCustomers = 1000;
GO


-- Create the "Marketing.GetLastThreeOrdersForSelectedCustomers" stored procedure with a temporary table

CREATE OR ALTER PROCEDURE
	Marketing.GetLastThreeOrdersForSelectedCustomers_TemporaryTable
(
	@NumberOfCustomers AS INT
)
AS

CREATE TABLE
	#SelectedCustomers
(
	CustomerId INT NOT NULL
);

INSERT INTO
	#SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

SELECT
	CustomerId		= SelectedCustomers.CustomerId ,
	OrderId			= LastThreeOrders.OrderId ,
	OrderDateTime	= LastThreeOrders.OrderDateTime ,
	OrderStatusId	= LastThreeOrders.OrderStatusId
FROM
	#SelectedCustomers AS SelectedCustomers
OUTER APPLY
	(
		SELECT TOP (3)
			OrderId			= Orders.Id ,
			OrderDateTime	= Orders.DateAndTime ,
			OrderStatusId	= Orders.OrderStatusId
		FROM
			Sales.Orders AS Orders
		WHERE
			Orders.CustomerId = SelectedCustomers.CustomerId
		ORDER BY
			OrderDateTime DESC
	)
	AS
		LastThreeOrders;

DROP TABLE
	#SelectedCustomers;
GO


-- Test the stored procedure
-- Notice that the estimated number of rows in the temporary table is 1,000

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TemporaryTable
	@NumberOfCustomers = 1000;
GO


-- Start an event session to track recompiles


-- Empty the plan cache and test the two stored procedures again
-- The table variable statements do not trigger any recompilations, while the temporary table statements trigger deferred compilations

DBCC FREEPROCCACHE;
GO


EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
	@NumberOfCustomers = 100;
GO


EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TemporaryTable
	@NumberOfCustomers = 100;
GO


-- Test the temporary table stored procedure 5 more times
-- Notice that there are no recompilations

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TemporaryTable
	@NumberOfCustomers = 100;
GO 5


-- Test the stored procedure with a different number of rows in the temporary table
-- Notice that there are still no recompilations

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TemporaryTable
	@NumberOfCustomers = 101;
GO


-- Test the stored procedure with a much different number of rows in the temporary table
-- Now there is a recompile due to statistics update

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TemporaryTable
	@NumberOfCustomers = 1000;
GO


-- Enable Table Variable Deferred Compilation
-- This feature is enabled by default for new databases in SQL Server 2019 and in Azure SQL

ALTER DATABASE SCOPED CONFIGURATION
SET DEFERRED_COMPILATION_TV = ON;
GO


-- Let's test the table variable stored procedures again
-- Now we get two deferred compilations, just like with the temporary table
-- Notice that the estimated number of rows is correct (100)
-- But since there are no statistics, the optimizer assumes there are 10 distinct customers,
-- and generates redundant Sort and Table Spool operations

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
	@NumberOfCustomers = 100;
GO


-- We can help the optimizer by adding a primary key constraint to the table variable

CREATE OR ALTER PROCEDURE
	Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
(
	@NumberOfCustomers AS INT
)
AS

DECLARE
	@SelectedCustomers
TABLE
(
	CustomerId INT NOT NULL PRIMARY KEY
);

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

SELECT
	CustomerId		= SelectedCustomers.CustomerId ,
	OrderId			= LastThreeOrders.OrderId ,
	OrderDateTime	= LastThreeOrders.OrderDateTime ,
	OrderStatusId	= LastThreeOrders.OrderStatusId
FROM
	@SelectedCustomers AS SelectedCustomers
OUTER APPLY
	(
		SELECT TOP (3)
			OrderId			= Orders.Id ,
			OrderDateTime	= Orders.DateAndTime ,
			OrderStatusId	= Orders.OrderStatusId
		FROM
			Sales.Orders AS Orders
		WHERE
			Orders.CustomerId = SelectedCustomers.CustomerId
		ORDER BY
			OrderDateTime DESC
	)
	AS
		LastThreeOrders;
GO


-- Let's test the stored procedures again
-- Now the optimizer estimated correctly that there are 100 distinct customers

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
	@NumberOfCustomers = 100;
GO


-- Test the stored procedure with a much different number of rows in the table variable
-- There is no statistics update in this case, so the plan remains the same, and the estimation is still 100 rows

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
	@NumberOfCustomers = 1000;
GO


-- In all the tests so far we didn't have an index on the CustomerId column in the Sales.Orders table
-- Let's create the index and see what happens...

CREATE NONCLUSTERED INDEX
	ix_Orders_nc_nu_CustomerId
ON
	Sales.Orders (CustomerId ASC);
GO


-- Test the table variable stored procedure again
-- Notice that the estimated number of rows in the table variable is now correct (1,000)
-- This is because there was a recompile due to a schema change (the new index)


EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
	@NumberOfCustomers = 1000;
GO


-- Test the two stored procedures again with 10,000 customers
-- Notice that in this case the optimizer was wrong, and the plan for the temporary table is worse due to the heavy use of the index spool

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
	@NumberOfCustomers = 10000;
GO


EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TemporaryTable
	@NumberOfCustomers = 10000;
GO


-- We can help the optimizer in various ways, such as adding the missing columns as included columns in the index

DROP INDEX
	ix_Orders_nc_nu_CustomerId
ON
	Sales.Orders;
GO


CREATE NONCLUSTERED INDEX
	ix_Orders_nc_nu_CustomerId_DateAndTime#OrderStatusId
ON
	Sales.Orders (CustomerId ASC)
INCLUDE
(
	DateAndTime ,
	OrderStatusId
);
GO


-- Test the two stored procedures again with 10,000 customers
-- Notice that now the two plans are identical

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TableVariable
	@NumberOfCustomers = 10000;
GO


EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_TemporaryTable
	@NumberOfCustomers = 10000;
GO


-- Create the "Marketing.GetLastThreeOrdersForSelectedCustomers" stored procedure with a table variable and OPTION (RECOMPILE)

CREATE OR ALTER PROCEDURE
	Marketing.GetLastThreeOrdersForSelectedCustomers_OptionRecompile
(
	@NumberOfCustomers AS INT
)
AS

DECLARE
	@SelectedCustomers
TABLE
(
	CustomerId INT NOT NULL PRIMARY KEY
);

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

SELECT
	CustomerId		= SelectedCustomers.CustomerId ,
	OrderId			= LastThreeOrders.OrderId ,
	OrderDateTime	= LastThreeOrders.OrderDateTime ,
	OrderStatusId	= LastThreeOrders.OrderStatusId
FROM
	@SelectedCustomers AS SelectedCustomers
OUTER APPLY
	(
		SELECT TOP (3)
			OrderId			= Orders.Id ,
			OrderDateTime	= Orders.DateAndTime ,
			OrderStatusId	= Orders.OrderStatusId
		FROM
			Sales.Orders AS Orders
		WHERE
			Orders.CustomerId = SelectedCustomers.CustomerId
		ORDER BY
			OrderDateTime DESC
	)
	AS
		LastThreeOrders
OPTION
	(RECOMPILE);
GO


-- Test the stored procedure
-- Notice the recompile event with the reason "Option (recompile) requested"
-- Also notice that the estimated number of rows is correct (1,000)

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_OptionRecompile
	@NumberOfCustomers = 1000;
GO


-- Test the stored procedure with a much different number of rows in the table variable
-- The statement is recompiled in each execution at runtime, so now the estimation is always correct

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_OptionRecompile
	@NumberOfCustomers = 10000;
GO


-- Create the "Marketing.CountOrdersForSelectedCustomers" stored procedure with a table-valued parameter

CREATE TYPE
	Marketing.SelectedCustomers
AS TABLE
(
	CustomerId INT NOT NULL
);
GO


CREATE OR ALTER PROCEDURE
	Marketing.CountOrdersForSelectedCustomers_TableParameter
(
	@SelectedCustomers AS Marketing.SelectedCustomers READONLY
)
AS

SELECT
	CustomerId		= SelectedCustomers.CustomerId ,
	OrderId			= LastThreeOrders.OrderId ,
	OrderDateTime	= LastThreeOrders.OrderDateTime ,
	OrderStatusId	= LastThreeOrders.OrderStatusId
FROM
	@SelectedCustomers AS SelectedCustomers
OUTER APPLY
	(
		SELECT TOP (3)
			OrderId			= Orders.Id ,
			OrderDateTime	= Orders.DateAndTime ,
			OrderStatusId	= Orders.OrderStatusId
		FROM
			Sales.Orders AS Orders
		WHERE
			Orders.CustomerId = SelectedCustomers.CustomerId
		ORDER BY
			OrderDateTime DESC
	)
	AS
		LastThreeOrders;
GO


-- Test the stored procedure with 1,000 rows
-- Notice that the estimated number of rows is correct (1,000) due to parameter sniffing in the first execution
-- But we still get the redundant sort and table spool operators due to lack of statistics (we removed the primary key)

DECLARE
	@SelectedCustomers	AS Marketing.SelectedCustomers ,
	@NumberOfCustomers	AS INT	= 1000;

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

EXECUTE Marketing.CountOrdersForSelectedCustomers_TableParameter
	@SelectedCustomers = @SelectedCustomers;
GO


-- Test the stored procedure with 10,000 rows
-- Notice that the estimated number of rows is still 1,000

DECLARE
	@SelectedCustomers	AS Marketing.SelectedCustomers ,
	@NumberOfCustomers	AS INT	= 10000;

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

EXECUTE Marketing.CountOrdersForSelectedCustomers_TableParameter
	@SelectedCustomers = @SelectedCustomers;
GO


-- One way to solve the problem is to copy the data from the table-valued parameter to a temporary table
-- But sometimes the overhead will be greater than the savings

CREATE OR ALTER PROCEDURE
	Marketing.CountOrdersForSelectedCustomers_TableParameter
(
	@SelectedCustomers AS Marketing.SelectedCustomers READONLY
)
AS

CREATE TABLE
	#SelectedCustomers
(
	CustomerId INT NOT NULL
);

INSERT INTO
	#SelectedCustomers
SELECT
	CustomerId
FROM
	@SelectedCustomers;

SELECT
	CustomerId		= SelectedCustomers.CustomerId ,
	OrderId			= LastThreeOrders.OrderId ,
	OrderDateTime	= LastThreeOrders.OrderDateTime ,
	OrderStatusId	= LastThreeOrders.OrderStatusId
FROM
	#SelectedCustomers AS SelectedCustomers
OUTER APPLY
	(
		SELECT TOP (3)
			OrderId			= Orders.Id ,
			OrderDateTime	= Orders.DateAndTime ,
			OrderStatusId	= Orders.OrderStatusId
		FROM
			Sales.Orders AS Orders
		WHERE
			Orders.CustomerId = SelectedCustomers.CustomerId
		ORDER BY
			OrderDateTime DESC
	)
	AS
		LastThreeOrders;
GO


-- Test the stored procedure with 1,000 rows
-- Notice that the estimated number of rows is correct (1,000) due to deferred compilation in the first execution
-- Also notice that thanks to statistics we got rid of the redundant sort and table spool operators

DECLARE
	@SelectedCustomers	AS Marketing.SelectedCustomers ,
	@NumberOfCustomers	AS INT	= 1000;

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

EXECUTE Marketing.CountOrdersForSelectedCustomers_TableParameter
	@SelectedCustomers = @SelectedCustomers;
GO


-- Test the stored procedure with 10,000 rows
-- Notice that the estimated number of rows is now correct due to statistics change, and the new plan is parallel

DECLARE
	@SelectedCustomers	AS Marketing.SelectedCustomers ,
	@NumberOfCustomers	AS INT	= 10000;

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

EXECUTE Marketing.CountOrdersForSelectedCustomers_TableParameter
	@SelectedCustomers = @SelectedCustomers;
GO


-- Another way to solve the problem is to use OPTION (RECOMPILE)
-- But if the execution frequency is too high, this can be dangerous

CREATE OR ALTER PROCEDURE
	Marketing.CountOrdersForSelectedCustomers_TableParameter
(
	@SelectedCustomers AS Marketing.SelectedCustomers READONLY
)
AS

SELECT
	CustomerId		= SelectedCustomers.CustomerId ,
	OrderId			= LastThreeOrders.OrderId ,
	OrderDateTime	= LastThreeOrders.OrderDateTime ,
	OrderStatusId	= LastThreeOrders.OrderStatusId
FROM
	@SelectedCustomers AS SelectedCustomers
OUTER APPLY
	(
		SELECT TOP (3)
			OrderId			= Orders.Id ,
			OrderDateTime	= Orders.DateAndTime ,
			OrderStatusId	= Orders.OrderStatusId
		FROM
			Sales.Orders AS Orders
		WHERE
			Orders.CustomerId = SelectedCustomers.CustomerId
		ORDER BY
			OrderDateTime DESC
	)
	AS
		LastThreeOrders
OPTION
	(RECOMPILE);
GO


-- Test the stored procedure with 1,000 rows
-- Notice that the estimated number of rows is correct (1,000) due to the OPTION (RECOMPILE)

DECLARE
	@SelectedCustomers	AS Marketing.SelectedCustomers ,
	@NumberOfCustomers	AS INT	= 1000;

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

EXECUTE Marketing.CountOrdersForSelectedCustomers_TableParameter
	@SelectedCustomers = @SelectedCustomers;
GO


-- Test the stored procedure with 10,000 rows
-- Notice that the estimated number of rows is still correct (10,000) due to the OPTION (RECOMPILE)

DECLARE
	@SelectedCustomers	AS Marketing.SelectedCustomers ,
	@NumberOfCustomers	AS INT	= 10000;

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

EXECUTE Marketing.CountOrdersForSelectedCustomers_TableParameter
	@SelectedCustomers = @SelectedCustomers;
GO


-- Sometimes the best way to solve the problem is to use trace flag 2453
-- This trace flag makes table variables behave like temporary tables in regards to statistics changes
-- But only the table cardinality (number of rows) is tracked; There is still no statistics

CREATE OR ALTER PROCEDURE
	Marketing.CountOrdersForSelectedCustomers_TableParameter
(
	@SelectedCustomers AS Marketing.SelectedCustomers READONLY
)
AS

DBCC TRACEON (2453);

SELECT
	CustomerId		= SelectedCustomers.CustomerId ,
	OrderId			= LastThreeOrders.OrderId ,
	OrderDateTime	= LastThreeOrders.OrderDateTime ,
	OrderStatusId	= LastThreeOrders.OrderStatusId
FROM
	@SelectedCustomers AS SelectedCustomers
OUTER APPLY
	(
		SELECT TOP (3)
			OrderId			= Orders.Id ,
			OrderDateTime	= Orders.DateAndTime ,
			OrderStatusId	= Orders.OrderStatusId
		FROM
			Sales.Orders AS Orders
		WHERE
			Orders.CustomerId = SelectedCustomers.CustomerId
		ORDER BY
			OrderDateTime DESC
	)
	AS
		LastThreeOrders;

DBCC TRACEOFF (2453);
GO


-- Test the stored procedure with 1,000 rows
-- Notice that the estimated number of rows is correct (1,000) due to deferred compilation

DECLARE
	@SelectedCustomers	AS Marketing.SelectedCustomers ,
	@NumberOfCustomers	AS INT	= 1000;

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

EXECUTE Marketing.CountOrdersForSelectedCustomers_TableParameter
	@SelectedCustomers = @SelectedCustomers;
GO


-- Test the stored procedure with 10,000 rows
-- Notice that the estimated number of rows is still correct (10,000) due to statistics change

DECLARE
	@SelectedCustomers	AS Marketing.SelectedCustomers ,
	@NumberOfCustomers	AS INT	= 10000;

INSERT INTO
	@SelectedCustomers
SELECT TOP (@NumberOfCustomers)
	CustomerId = Id
FROM
	Marketing.Customers
ORDER BY
	NEWID () ASC;

EXECUTE Marketing.CountOrdersForSelectedCustomers_TableParameter
	@SelectedCustomers = @SelectedCustomers;
GO


-- Sometimes the best solution is to use a hybrid approach

CREATE OR ALTER PROCEDURE
	Marketing.GetLastThreeOrdersForSelectedCustomers_Hybrid
(
	@NumberOfCustomers AS INT
)
AS

IF
	@NumberOfCustomers <= 10000
BEGIN

	DECLARE
		@SelectedCustomers
	TABLE
	(
		CustomerId INT NOT NULL PRIMARY KEY
	);

	INSERT INTO
		@SelectedCustomers
	SELECT TOP (@NumberOfCustomers)
		CustomerId = Id
	FROM
		Marketing.Customers
	ORDER BY
		NEWID () ASC;

	SELECT
		CustomerId		= SelectedCustomers.CustomerId ,
		OrderId			= LastThreeOrders.OrderId ,
		OrderDateTime	= LastThreeOrders.OrderDateTime ,
		OrderStatusId	= LastThreeOrders.OrderStatusId
	FROM
		@SelectedCustomers AS SelectedCustomers
	OUTER APPLY
		(
			SELECT TOP (3)
				OrderId			= Orders.Id ,
				OrderDateTime	= Orders.DateAndTime ,
				OrderStatusId	= Orders.OrderStatusId
			FROM
				Sales.Orders AS Orders
			WHERE
				Orders.CustomerId = SelectedCustomers.CustomerId
			ORDER BY
				OrderDateTime DESC
		)
		AS
			LastThreeOrders;

END
ELSE	-- @NumberOfCustomers > 10000
BEGIN

	CREATE TABLE
		#SelectedCustomers
	(
		CustomerId INT NOT NULL
	);

	INSERT INTO
		#SelectedCustomers
	SELECT TOP (@NumberOfCustomers)
		CustomerId = Id
	FROM
		Marketing.Customers
	ORDER BY
		NEWID () ASC;

	SELECT
		CustomerId		= SelectedCustomers.CustomerId ,
		OrderId			= LastThreeOrders.OrderId ,
		OrderDateTime	= LastThreeOrders.OrderDateTime ,
		OrderStatusId	= LastThreeOrders.OrderStatusId
	FROM
		#SelectedCustomers AS SelectedCustomers
	OUTER APPLY
		(
			SELECT TOP (3)
				OrderId			= Orders.Id ,
				OrderDateTime	= Orders.DateAndTime ,
				OrderStatusId	= Orders.OrderStatusId
			FROM
				Sales.Orders AS Orders
			WHERE
				Orders.CustomerId = SelectedCustomers.CustomerId
			ORDER BY
				OrderDateTime DESC
		)
		AS
			LastThreeOrders;

	DROP TABLE
		#SelectedCustomers;

END;
GO


-- Test the stored procedure with 1,000 rows
-- The estimated number of rows is correct (1,000) due to deferred compilation

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_Hybrid
	@NumberOfCustomers = 1000;
GO


-- Test the stored procedure with 10,000 rows
-- The estimated number of rows is still 1,000

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_Hybrid
	@NumberOfCustomers = 10000;
GO


-- Test the stored procedure with 100,000 rows
-- The estimated number of rows is correct (100,000) due to deferred compilation

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_Hybrid
	@NumberOfCustomers = 100000;
GO


-- Test the stored procedure with 50,000 rows
-- The estimated number of rows is correct (50,000) due to statistics update

EXECUTE Marketing.GetLastThreeOrdersForSelectedCustomers_Hybrid
	@NumberOfCustomers = 50000;
GO
